(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2021 *)

(** Type definitions for the abstract syntax tree of a simple 'while'
   language with references and function calls, whose programs are
   written in administrative normal form. *)

type var = string
(** Program variables *)

type loc = int
(** Memory locations, a.k.a addresses *)

type fname = string
(** Function names *)

(** Values *)
type v =
  | Unit  (** Unit value *)
  | Bool of bool  (** Boolean value *)
  | Int of int  (** Integer value *)
  | Loc of loc  (** Memory location *)

(** Binary operators *)
type binop =
  | AND  (** Boolean conjunction *)
  | OR  (** Boolean disjunction *)
  | LE  (** Lesser than or equal integer test *)
  | LT  (** Lesser than integer test *)
  | GE  (** Greater than or equal integer test *)
  | GT  (** Greater than integer test *)
  | NEQ  (** Disequality integer test *)
  | EQ  (** Equality integer test *)
  | PLUS  (** Integer addition *)
  | MINUS  (** Integer subtraction *)
  | MULT  (** Integer multiplication *)

(** Immediate terms *)
type imm = Val of v  (** Values *) | Var of var  (** Variables *)

(** Atomic terms *)
type a =
  | Imm of imm  (** Immediate values *)
  | Binop of binop * imm * imm  (** Binary operation *)
  | NewRef of imm
      (** Creation of a reference, initialized with the value denoted by the
          immediate that is passed as argument *)
  | GetRef of var  (** Read of the contents of a reference *)
  | SetRef of var * imm  (** Write of the contents of a reference *)
  | Call of fname * imm list  (** Function call *)

(** Commands *)
type cmd =
  | Return of a  (** Atomic term *)
  | AssignVar of var * a
  | If of imm * cmds * cmds  (** If-Then-Else branching construct *)
  | While of imm * cmds  (** While loop *)

and cmds = cmd list
(** Sequence of commands are non-empty lists of commands *)

type func = var list * cmds
(** A function is a list of formal arguments and a term that
   represents its code *)

type program = (fname * func) list
(** A program is a list of function declarations *)
