(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2022 *)

module Check = struct
  exception E of string

  module StringMap = Map.Make (String)
  module StringSet = Set.Make (String)

  let immediate vars = function
    | Ast.Val _ -> ()
    | Var x ->
        if StringSet.mem x vars then () else raise (E ("Unknown variable " ^ x))

  let atomic arities vars = function
    | Ast.Imm i -> immediate vars i
    | Binop (_binop, i1, i2) ->
        immediate vars i1;
        immediate vars i2
    | NewRef i -> immediate vars i
    | GetRef x -> immediate vars (Ast.Var x)
    | SetRef (x, i) ->
        immediate vars (Ast.Var x);
        immediate vars i
    | Call (f, args) -> (
        match StringMap.find_opt f arities with
        | None -> raise (E ("unknown function " ^ f))
        | Some n ->
            let m = List.length args in
            if n <> m then
              raise
                (E
                   (Format.sprintf
                      "function %s expects %i arguments, but %i arguments were \
                       provided"
                      f n m))
            else List.iter (immediate vars) args)

  let rec cmd arities vars = function
    | Ast.Return a ->
        atomic arities vars a;
        vars
    | AssignVar (x, a) ->
        atomic arities vars a;
        StringSet.add x vars
    | If (i, cmds1, cmds2) ->
        immediate vars i;
        let vars1 = cmds arities vars cmds1
        and vars2 = cmds arities vars cmds2 in
        StringSet.union vars1 vars2
    | While (i, b) ->
        immediate vars i;
        cmds arities vars b

  and cmds arities xs b = List.fold_left (cmd arities) xs b

  let args xs =
    List.fold_left
      (fun acc x ->
        if StringSet.mem x acc then
          raise (E ("argument " ^ x ^ " is declared more than once."))
        else StringSet.add x acc)
      StringSet.empty xs

  let decl arities (f, (xs, b)) =
    try
      let vars = args xs in
      ignore @@ cmds arities vars b
    with E msg -> raise (E ("In function " ^ f ^ ":\n" ^ msg))

  let program (ds : Ast.program) =
    try
      let arities =
        List.fold_left
          (fun acc (name, (args, _body)) ->
            if StringMap.mem name acc then
              raise (E ("Function " ^ name ^ " is declared more than once."))
            else StringMap.add name (List.length args) acc)
          StringMap.empty ds
      in
      List.iter (decl arities) ds;
      Ok ()
    with E msg -> Error ("Error:\n" ^ msg)
end

(** Performs some syntactic checks on a program *)
let check = Check.program
