(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2019
*)

(** Driver for lexing/parsing *)

let chunk input (pos1, pos2) =
  let start = pos1.Lexing.pos_cnum in
  let len = pos2.Lexing.pos_cnum - start in
  match input with
  | `String s -> String.sub s start len
  | `Channel chn ->
      let old_pos = pos_in chn in
      seek_in chn start;
      let s = really_input_string chn len in
      seek_in chn old_pos;
      s

let show input poss =
  chunk input poss |> MenhirLib.ErrorReports.sanitize
  |> MenhirLib.ErrorReports.compress
  |> MenhirLib.ErrorReports.shorten 20

let fail input locs_buffer lexbuf (s : int) =
  (* Display a nice error message. In principle, the table found in
     [ParserMessages] should be complete, so we should obtain
     a nice message. If [Not_found] is raised, we produce a generic
     message, which is better than nothing. Note that the OCaml code
     in [ParserMessages] is auto-generated based on the table in
     [ParserMessages.messages]. *)
  let message =
    try ParserMessages.message s
    with Not_found -> Printf.sprintf "Unknown syntax error (in state %d).\n" s
  in
  (* Show the two tokens between which the error took place. *)
  let where = MenhirLib.ErrorReports.show (show input) locs_buffer in
  (* Hack: remove the final newline, because [Error.error] adds one. *)
  let message = String.sub message 0 (String.length message - 1) in
  (* Build the error message message. *)
  Format.asprintf "%a@.Syntax error %s.\n%s" Location.pp (Location.get lexbuf)
    where message

(** [parse parser name input] parses the input [input] named [name]
   using the parsing function [parser] *)
let parse parser name input =
  let lexbuf =
    match input with
    | `String s -> Lexing.from_string s
    | `Channel chn -> Lexing.from_channel chn
  in
  Lexing.set_filename lexbuf name;
  let locs_buffer, lexer = MenhirLib.ErrorReports.wrap Lexer.token in
  match parser lexer lexbuf with
  | ast -> Ok ast
  | exception Lexer.Error (loc, token) ->
      let msg =
        Format.asprintf "%a@.Error: unexpected character %s" Location.pp loc
          token
      in
      Error msg
  | exception Parser.Error s ->
      let msg = fail input locs_buffer lexbuf s in
      Error msg

(** [parse_file parser file] parses the file [file] using the parsing
   function [parser] *)
let parse_file parser filename =
  let chn = open_in filename in
  let res = parse parser filename (`Channel chn) in
  close_in chn;
  res
